/*
 * Copyright 1995 Steve Passe
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of AUTHOR (Steve Passe)
 * is not be used in advertising or publicity pertaining to distribution of
 * the software without specific, written prior permission.  AUTHOR makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHOR
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * AUTHOR:     Steve Passe <smp@csn.net>
*/

#define DEBUG
#define FLATWHEELS


/*
 * xadc.c, monitor ADC controller
 */

static char rcsid[] = "$Id: xadc.c,v 1.4 1995/09/15 09:24:14 smp Exp smp $";

#include <stdio.h>

/* Xlib headers */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

/* Standard Xt headers */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

/* FWF Widget Set headers */
#include <Xfwf/ThWheel.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>

/* Athena Widget Set headers */
#include <X11/Xaw/Form.h>

/* Control Panel Widget Set headers */
#include <X11/Xc/Xc.h>
#include <X11/Xc/Lcd2.h>

#include "xadc.h"

/* Widgets */
Widget TopLevel;
Widget		MainBox;
display_t		gauge[ MAX_CHANNELS ];


static String fallback_resources[] = {

    ".channelCount:			2",
    ".channelList:			",

    /* per channel stuff */
    "*Form.Label.background:		" HI_BACKGROUND,
#if defined( FLATWHEELS )
    "*Form.Label.width:			250",
#else
    "*Form.Label.width:			215",
#endif

    /* lcd displays */
    "*Lcd2.shadeDepth:			2",
    "*Lcd2.background:			" LO_BACKGROUND,
    "*Lcd2.valueForeground:		orange",
    "*Lcd2.valueBackground:		black",
    "*Lcd2.width:			150",
    "*Lcd2.height:			65",
    "*Lcd2.dataType:			lval",
    "*Lcd2.upperBound:			300",
    "*Lcd2.lowerBound:			40",
    "*Lcd2.value:			60",
    "*Lcd2.valueLength:			3",
    "*Lcd2.interval:			30000",		/* 30 seconds */

    /* WARN/PANIC formes */
#if defined( FLATWHEELS )
    "*Form.Form.Label.internalHeight:	1",
#endif
    "*Form.Form.Label.width:		29",
    "*Form.Form.Label.value:		---",
    "*Form.Form.Label.background:	" HI_BACKGROUND,
/**    "*XfwfThumbWheel.vertical:		False", */
    "*XfwfThumbWheel.minValue:		-1",
    "*XfwfThumbWheel.maxValue:		1000",
    "*XfwfThumbWheel.value:		-1",
    "*XfwfThumbWheel.step:		1",

    /* all others */
    "*background:			" DEFAULT_BACKGROUND,
    NULL,
};


static int channelCount = 1;

typedef struct {
	int	channels;
	char	*list;
} appResources_t, *appResourcesPtr;

static appResources_t	appResources;

static XtResource	resources[] = {
#define offset(field)		XtOffset(appResourcesPtr, field)
	{ XtNchannelCount, XtCChannelCount, XtRInt, sizeof(int), 
		offset(channels), XtRImmediate, (XtPointer) 2 },
	{ XtNchannelList, XtCChannelList, XtRString, sizeof(char *),
		offset(list), XtRImmediate, "" },
#undef offset
};

static XrmOptionDescRec options[] = {
	{ "-c",	".channelCount",	XrmoptionSepArg, (caddr_t) 2 },
	{ "-l",	".channelList",		XrmoptionSepArg, (caddr_t) "" },
};


/* private functions */
static void LcdCallback( Widget, XtPointer, XcData );
static void WheelWarnCallback( Widget, XtPointer, XtPointer );
static void WheelPanicCallback( Widget, XtPointer, XtPointer );
static void WheelRateCallback( Widget, XtPointer, XtPointer );

int getSample( int channel );

static void
buildAdj( char* name, int instance,
	  adjustor_t* ap,
	  Widget parent, Widget ftop, Widget fleft );

/*
 *
 */
void
main( int argc, String *argv )
{
    XtAppContext app_context;
    Display *display;
    Arg args[ 10 ];
    char s[ 133 ];
    int value;
    int list[ MAX_CHANNELS + 1 ];
    int x, xx;
    int i;

    /* create top level shell */
    TopLevel = XtVaAppInitialize(&app_context, "Xadc", options, 2, &argc, argv,
				 fallback_resources, NULL);

    /* process remaining commandline args */
    XtGetApplicationResources(TopLevel, &appResources, resources,
			      XtNumber(resources), 0, 0);

    if ( (channelCount = appResources.channels) > MAX_CHANNELS )
    {
	channelCount = MAX_CHANNELS;
    }
    
    if ( appResources.list && *appResources.list )
    {
	x = sscanf( appResources.list, "%d,%d,%d,%d,%d,%d,%d,%d",
	        &list[0], &list[1], &list[2], &list[3],
	        &list[4], &list[5], &list[6], &list[7] );
	if ( x > MAX_CHANNELS )
	{
	    fprintf( stderr, "bad channel list\n" );
	    exit(1);
	}

	channelCount = x;
	while ( x < MAX_CHANNELS )
	{
	    list[x++] = NULL;
	}
    }
    else
    {
	for ( x = 0; x < channelCount; ++x )
	{
	    list[ x ] = x+1;
	}
    }

#if defined( DEBUG )
    fprintf( stderr, "channels:" );
    for ( x = 0; x < MAX_CHANNELS; ++x )
    {
	if ( list[x] ) fprintf( stderr, " %d", list[x] );
    }
    fprintf( stderr, "\n" );
#endif /* DEBUG */

    /* general X stuff */
    display = XtDisplay(TopLevel);
 
    /* set resources for the toplevel shell */
#if 0
    i = 0;
    XtSetArg(args[i], XtNminWidth, 250);  i++;
    XtSetValues(TopLevel, args, i);
#endif

    /* the top level container */
    MainBox = XtVaCreateManagedWidget("mainbox", boxWidgetClass,
				      TopLevel, NULL);
    i = 0;
    XtSetArg(args[i], XtNdefaultDistance, 0);  i++;
    XtSetValues(MainBox, args, i);

    i = 0;
    XtSetArg(args[i], XtNvSpace, 0); i++;
    XtSetArg(args[i], XtNhSpace, 0); i++;
    XtSetValues(MainBox, args, i);


    /* create forms, lcd displays, wheels, labels, etc. */
    for ( xx = 0; xx < channelCount; ++xx )
    {
	/* get indexes from ordered list */
	x = list[xx]-1;

	/*
	 * outer form
	 */
	sprintf( s, "form_%d", x+1 );
	gauge[ x ].form = XtVaCreateManagedWidget(s, formWidgetClass,
					   MainBox, NULL);
	i = 0;
	XtSetArg(args[i], XtNvertDistance, 0); i++;
	XtSetArg(args[i], XtNhorizDistance, 0); i++;
	XtSetArg(args[i], XtNdefaultDistance, 0); i++;
	XtSetValues(gauge[ x ].form, args, i);

	/*
	 * channel label
	 */
	sprintf( s, "channel_%d", x+1 );
	gauge[ x ].label = XtVaCreateManagedWidget(s, labelWidgetClass,
					     gauge[ x ].form, NULL);

	/*
	 * channel display
	 */
	sprintf( s, "lcd_%d", x+1 );
	gauge[ x ].lcd = XtVaCreateManagedWidget(s, xcLcd2WidgetClass,
					 gauge[ x ].form, NULL);
	i = 0;
	XtSetArg(args[i], XtNfromVert, gauge[ x ].label); i++;
	XtSetValues(gauge[ x ].lcd, args, i);

	/*
	 * build WARN form
	 */
	buildAdj( "warn", x, &gauge[x].warn,
		  gauge[x].form, gauge[x].label, gauge[x].lcd );

	WheelWarnCallback( gauge[ x ].warn.wheel, (XtPointer)x, NULL );

	/*
	 * build PANIC form
	 */
	buildAdj( "panic", x, &gauge[x].panic,

#if defined( FLATWHEELS )
		  gauge[x].form, gauge[x].warn.form, gauge[x].lcd );
#else
		  gauge[x].form, gauge[x].label, gauge[x].warn.form );
#endif

	WheelPanicCallback( gauge[ x ].panic.wheel, (XtPointer)x, NULL );

	/*
	 * build RATE form
	 */
	buildAdj( "rate", x, &gauge[x].rate,

#if defined( FLATWHEELS )
		  gauge[x].form, gauge[x].panic.form, gauge[x].lcd );
#else
		  gauge[x].form, gauge[x].label, gauge[x].panic.form );
#endif

	WheelRateCallback( gauge[ x ].rate.wheel, (XtPointer)x, NULL );

	/*
	 * register callback functions
	 */
	XtAddCallback( gauge[x].lcd, XcNupdateCallback,
		      (XtCallbackProc)LcdCallback, (XtPointer)x );
	XtAddCallback( gauge[x].warn.wheel, XtNscrollCallback,
		      (XtCallbackProc)WheelWarnCallback, (XtPointer)x );
	XtAddCallback( gauge[x].panic.wheel, XtNscrollCallback,
		      (XtCallbackProc)WheelPanicCallback, (XtPointer)x );
	XtAddCallback( gauge[x].rate.wheel, XtNscrollCallback,
		      (XtCallbackProc)WheelRateCallback, (XtPointer)x );
    }

    XtRealizeWidget(TopLevel);
    XtAppMainLoop(app_context);
}


static void
buildAdj( char* name, int instance,
	  adjustor_t* ap,
	  Widget parent, Widget ftop, Widget fleft )
{
    Arg args[ 10 ];
    int i;
    char s[ 64 ];

    ++instance;

    sprintf( s, "%sA_%d", name, instance );
    ap->form = XtVaCreateManagedWidget(s, formWidgetClass, parent, NULL);

    /* define placement of container */
    i = 0;
    XtSetArg( args[i], XtNvertDistance, 0 ); i++;
    XtSetArg( args[i], XtNhorizDistance, 0 ); i++;
    XtSetArg( args[i], XtNdefaultDistance, 0 ); i++;
    XtSetArg( args[i], XtNfromVert, ftop ); i++;
    XtSetArg( args[i], XtNfromHoriz, fleft ); i++;
    XtSetValues( ap->form, args, i );

    /* create adjustor label */
    sprintf( s, "%sL_%d", name, instance );
    ap->label = XtVaCreateManagedWidget( s, labelWidgetClass,
					 ap->form, NULL );

    /* create adjustor wheel */
    sprintf( s, "%sW_%d", name, instance );
    ap->wheel = XtVaCreateManagedWidget( s, xfwfThumbWheelWidgetClass,
					 ap->form, NULL );

    /* plug default values into structure */
    XtVaGetValues( ap->wheel, XtNminValue, &(ap->min), NULL );
    XtVaGetValues( ap->wheel, XtNmaxValue, &(ap->max), NULL );
    XtVaGetValues( ap->wheel, XtNvalue, &(ap->val), NULL );

    i = 0;
#if defined( FLATWHEELS )
    XtSetArg( args[i], XtNfromHoriz, ap->label ); i++;
#else
    XtSetArg( args[i], XtNfromVert, ap->label ); i++;
#endif
    XtSetValues( ap->wheel, args, i );
}


/*
 * updateCallback for the Lcd widgets.  Simply increments the current value
 * based on the dataType.  Used to implement a simple count-up counter.
 */
static
void LcdCallback( Widget w, XtPointer client_data, XcData call_data )
{
    int x, warn, panic;
    char s[ 64 ];
    XcVType value;

    x =(int)client_data;

    /*
     * determine if WARNing or PANIC needed
     */
    value.lval = (long)getSample( x );
    XtVaGetValues( gauge[ x ].warn.wheel, XtNvalue, &warn, NULL );
    if ( value.lval >= warn )
    {
	warn = 1;
	XtVaGetValues( gauge[ x ].panic.wheel, XtNvalue, &panic, NULL );
	if ( value.lval >= panic )
	{
	    strcpy( s, PANIC_FCOLOR );
	    panic = 1;
	}
	else
	{
	    strcpy( s, WARN_FCOLOR );
	    panic = 0;
	}
    }
    else
    {
	strcpy( s, OK_FCOLOR );
	warn = panic = 0;
    }

    /* update lcd color */
    XtVaSetValues( gauge[ x ].lcd,
		  XtVaTypedArg, XcNvalueForeground, XtRString, s, sizeof(s)+1,
		  NULL );

    /* return the new value to lcd */
    call_data->value.lval=value.lval; 

    /*****
     * process warn & panic
     */

}


static
void WheelWarnCallback( Widget w, XtPointer client_data, XtPointer call_data )
{
    int x = (int)client_data;
    int value;
    char s1[64];
    char s2[64];

    /* get new value of wheel */
    XtVaGetValues( w, XtNvalue, &value, NULL );

    if ( (value > gauge[ x ].warn.min) && (value < gauge[ x ].warn.max) )
    {
	sprintf( s1, "%03d", value );
	strcpy( s2, WARN_COLOR );
    }
    else
    {
	strcpy( s1, NIL_LABEL );
	strcpy( s2, NIL_COLOR );
    }

    /* update label */
    XtVaSetValues( gauge[ x ].warn.label,
		  XtNlabel, s1,
		  XtVaTypedArg, XtNbackground, XtRString, s2, sizeof(s2)+1,
		  NULL );
}


static
void WheelPanicCallback( Widget w, XtPointer client_data, XtPointer call_data )
{
    int x = (int)client_data;
    int value;
    char s1[64];
    char s2[64];


    XtVaGetValues( w, XtNvalue, &value, NULL );

    if ( (value > gauge[ x ].warn.min) && (value < gauge[ x ].panic.max) )
    {
	sprintf( s1, "%03d", value );
	strcpy( s2, PANIC_COLOR );
    }
    else
    {
	strcpy( s1, NIL_LABEL );
	strcpy( s2, NIL_COLOR );
    }

    /* update label */
    XtVaSetValues( gauge[ x ].panic.label,
		  XtNlabel, s1,
		  XtVaTypedArg, XtNbackground, XtRString, s2, sizeof(s2)+1,
		  NULL );
}


static
void WheelRateCallback( Widget w, XtPointer client_data, XtPointer call_data )
{
    int x = (int)client_data;
    int value;
    char s1[64];
    char s2[64];

    /* get new value of wheel */
    XtVaGetValues( w, XtNvalue, &value, NULL );

    if ( (value > gauge[ x ].rate.min) && (value < gauge[ x ].rate.max) )
    {
	sprintf( s1, "%03d", value );
	strcpy( s2, RATE_COLOR );
    }
    else
    {
	strcpy( s1, NIL_LABEL );
	strcpy( s2, NIL_COLOR );
    }

    XtVaSetValues( gauge[ x ].rate.label,
		  XtNlabel, s1,
		  XtVaTypedArg, XtNbackground, XtRString, s2, sizeof(s2)+1,
		  NULL );

    /* update sample rate */
    /** TODO: make this a log scale in software */
    value *= 1000;
    XtVaSetValues( gauge[ x ].lcd,
		  XcNinterval, value,
		  NULL );
}


int
getSample( int channel )
{
    int r;

    r = rand() & 0x1f;			/* 0 -> 31 */
    r -= 15;				/* -15 -> 16 */
    r += 100;

    return r;
}


#if !defined( XADC )

/* Callback for the Quit button: simply quits the application. */
void Quit(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtDestroyWidget(TopLevel);
   exit(0);
}


/* Callback for the Reset button: resets Lcd1's counter value to 0. */
void Reset_lcd(w, lcd, call_data)
Widget w;
Widget lcd; 
caddr_t call_data;
{
XcVType value;

   value.lval = 0L;
   XcLcdUpdateValue(lcd, &value);

}

#endif /* !XADC */
